#!/usr/bin/python3
# --
# Copyright (C) 2019 it-novum GmbH, http://www.it-novum.com
# --
# Authors: Markus Stollwerk
# --
# License ...
# --

from suds.client import Client
from pprint import pprint
from sys import exit
import argparse
import json
import re
import os
import time
import sys

parser = argparse.ArgumentParser()
optional = parser._action_groups.pop()
required = parser.add_argument_group('required arguments')
required.add_argument('--hostname', help='hostnamae or ipadresse', required=True)
required.add_argument('--port', help='SAP Webservice Port 5XX13', required=True)
required.add_argument('--mode', help='[checkitem, gettree, workprocess, checksyslog, getsyslog]', required=True)
optional.add_argument('--username', help='[username to check with]')
optional.add_argument('--password', help='[password to check with]')
optional.add_argument('--mode2', help='[list, invertthresholds, regex]')
optional.add_argument('--warning', help='warning threshold')
optional.add_argument('--critical', help='critical threshold')
optional.add_argument('--item', help='item to check')
optional.add_argument('--statearg', help='Parameter to store last check timestamp per syslog check')
parser._action_groups.append(optional) # added this line
args=parser.parse_args()

hostname=args.hostname
port=args.port
mode=args.mode
mode2=args.mode2
warning=args.warning
critical=args.critical
item=args.item
username=args.username
password=args.password
statesyslog=args.statearg

RC=0
RCtxt=""
RCok=0
RCwarning=1
RCcritical=2
RCunknown=3
maxage=900
delimiter="##"

url="http://" + hostname  + ":" + port  + "/?wsdl"
tmpfile="/var/tmp/check_sap_webservices/" + hostname + "-" + port + ".tmp"
tmpfilesyslog="/var/tmp/check_sap_webservices/" + hostname + "-" + port + ".syslog.tmp"

if mode == "checksyslog":
    try:
        lastsyslog="/var/tmp/check_sap_webservices/" + hostname + "-" + port + ".syslog." + statesyslog + ".time.tmp"
    except:
        print("UNKNOWN- please provide statearg")
        exit(RCunknown)



if not os.path.exists("/var/tmp/check_sap_webservices/"):
    os.makedirs("/var/tmp/check_sap_webservices/")


tmpitems={}
resultitems={}

def GetAlertTree():
    try:
        client = Client(url, cache=None)
        result = client.service.GetAlertTree()
        result = result[0]
    except Exception as e:
        print("CRITICAL - could not connect to host %s on port %s: %s" % (hostname, port, str(e)))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)

    for counter in range(len(result)):
        tmpdict={}
        if result[counter].name is None:
            tmpdict["name"]="SAP"
        else:
            tmpdict["name"]=result[counter].name
        tmpdict["parent"]=result[counter].parent
        tmpitems[counter]=tmpdict

        path=""
        parent=result[counter].parent
        while parent > -1:
            path=tmpitems[parent]["name"] + delimiter + path
            parent=tmpitems[parent]["parent"]
        if result[counter].name is None:
            path=path + "SAP"
        else:
            path=path + result[counter].name

        tmpresult={}
        tmpresult["state"]=result[counter].ActualValue
        tmpresult["output"]=result[counter].description
        tmpresult["tid"]=result[counter].Tid
        resultitems[path]=tmpresult

    if mode2 == "list":
        pprint(resultitems)
    else:
        with open(tmpfile, 'w') as file:
            file.write(json.dumps(resultitems))

    print("OK - data receive")
    exit(RCok)

def ABAPReadSyslog():
    try:
        client = Client(url, cache=None)
        result = client.service.ABAPReadSyslog()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)


    for counter in range(len(result)):

        tmpresult={}
        tmpresult["time"]=time.mktime(time.strptime(result[counter].Time, "%Y %m %d %H:%M:%S"))
        tmpresult["user"]=result[counter].User
        tmpresult["client"]=result[counter].Client
        tmpresult["text"]=result[counter].Text
        tmpresult["state"]=result[counter].Severity
        resultitems[counter]=tmpresult

    with open(tmpfilesyslog, 'w') as file:
        file.write(json.dumps(resultitems))

    print("OK - data receive")
    exit(RCok)

def GetAlerts(tid):
    try:
        client = Client(url, cache=None)
        result = client.service.GetAlerts(tid)
        result=result[1][0]
        longoutput=""
        for counter in range(len(result)):
            longoutput=longoutput + result[counter].Time + " --- " + result[counter].Description + "\n"
    except:
        longoutput=""

    return longoutput

def GetValue(argitem):
    try:
        with open(tmpfile, 'r') as file:
            resultitems=json.loads(file.read())
    except:
        print("UNKNOWN - could not open tmpfile or its empty " + tmpfile)
        exit(RCunknown)

    if (time.time() - os.stat(tmpfile).st_mtime) > maxage:
        print("UNKNOWN - last data receive is to old %ds > %ds" % ((time.time() - os.stat(tmpfile).st_mtime), maxage))
        exit(RCunknown)
    for key in resultitems:
        if key.endswith(argitem):
            return resultitems[key]

def GetSyslogValue():

    returnvalues={}
    counttmp=0
    lasttime=""
    try:
        with open(tmpfilesyslog, 'r') as file:
            resultitems=json.loads(file.read())
    except:
        print("UNKNOWN - could not open tmpfile or its empty " + tmpfilesyslog)
        exit(RCunknown)

    if (time.time() - os.stat(tmpfilesyslog).st_mtime) > maxage:
        print("UNKNOWN - last data receive is to old %ds > %ds" % ((time.time() - os.stat(tmpfilesyslog).st_mtime), maxage))
        exit(RCunknown)


    try:
        with open(lastsyslog, 'r') as file:
            lasttime=float(file.read())
    except:
        with open(lastsyslog, 'w') as file:
            file.write(str(time.time()))
            return None


    for counter in range(len(resultitems)):
        if resultitems[str(counter)]["time"] > lasttime:
            returnvalues[counttmp]=resultitems[str(counter)]
            counttmp=counttmp+1

    with open(lastsyslog, 'w') as file:
        file.write(str(resultitems["0"]["time"]))
    return(returnvalues)

def ABAPGetWPTable(argitem):

    try:
        client = Client(url, cache=None)
        result = client.service.ABAPGetWPTable()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)

    count=0
    for counter in range(len(result)):
            if str(result[counter].Typ) == argitem and result[counter].Status == "Wait":
                count=count+1

    return count

def NagiosOut(argvalues):

    if mode == "checkitem":
        if not delimiter in item:
            print("UNKNOWN - delimiter %s not found in item %s please submit at least one parent like parent##metric" % (delimiter, item))
            exit(RCunknown)
        vitem=item.replace(delimiter, " ")
        try:
            vstate=argvalues["state"]
            voutput=argvalues["output"]
            voutputlong=argvalues["output"]
            vtid=argvalues["tid"]
            if voutput is not None:
                try:
                    voutput=float(re.findall('[0-9]+', voutput)[0])
                except:
                    voutput=None
            else:
                voutputlong=vstate
        except:
            print("UNKNOWN - could not find " + item)
            exit(RCunknown)

        lout=GetAlerts(vtid)


        if warning is not None and critical is not None and warning != "na" and critical != "na":
            warn=float(re.findall('[0-9]+', warning)[0])
            crit=float(re.findall('[0-9]+', critical)[0])
            if mode2 == "invertthresholds":
                if warn < crit:
                    print("UNKNOWN - in mode invertthresholds warning should be larger then critical")
                    exit(RCunknown)

                if voutput > warn:
                    RCtxt="OK"
                    RC=RCok
                elif voutput < crit:
                    RCtxt="CRITICAL"
                    RC=RCcritical
                else:
                    RCtxt="WARNING"
                    RC=RCwarning
                print("%s - %s is %s|%s=%s" % (RCtxt, vitem, voutputlong, vitem.replace(' ', '_'), voutput))
            else:
                if warn > crit:
                    print("UNKNOWN - warning should be smaller then critical")
                    exit(RCunknown)

                if voutput < warn:
                    RCtxt="OK"
                    RC=RCok
                elif voutput > crit:
                    RCtxt="CRITICAL"
                    RC=RCcritical
                else:
                    RCtxt="WARNING"
                    RC=RCwarning
                print("%s - %s is %s|%s=%s;%s;%s" % (RCtxt, vitem, voutputlong, vitem.replace(' ', '_'), voutput, warn, crit))

        else:
            if "GREEN" in vstate:
                RCtxt="OK"
                RC=RCok
            elif "YELLOW" in vstate:
                RCtxt="WARNING"
                RC=RCwarning
            elif "RED" in vstate:
                RCtxt="CRITICAL"
                RC=RCcritical
            elif "GRAY" in vstate:
                RCtxt="UNKNOWN"
                RC=RCunknown
            print("%s - %s is %s|%s=%s" % (RCtxt, vitem, voutputlong, vitem.replace(' ', '_'), voutput))
        if len(lout) > 0:
            print(lout)
        exit(RC)

    elif mode == "workprocess":
        voutput=int(argvalues)
        try:
            warn=int(re.findall('[0-9]+', warning)[0])
            crit=int(re.findall('[0-9]+', critical)[0])
        except:
            print("UNKNOWN - please submit thresholds")
            exit(RCunknown)

        if crit > warn:
            print("UNKNOWN - critical threshold should be less then warning")

        if voutput > warn:
            RCtxt="OK"
            RC=RCok
        elif voutput <= crit:
            RCtxt="CRITICAL"
            RC=RCcritical
        else:
            RCtxt="WARNING"
            RC=RCwarning

        print("%s - there are %s %s workprocess in state Wait|%s=%s" % (RCtxt, voutput, item, item, voutput))
        exit(RC)

    elif mode == "checksyslog":
        errentries=""
        errentrieswarn=""
        errentriescrit=""
        RC=RCok
        if argvalues is None:
            print("OK - first run. Wait for the next check")
            exit(RCok)
        elif len(argvalues) == 0:
            print("OK - no new entries since last check")
            exit(RCok)
        else:
            if mode2 == "regex":
                if item is None:
                    print("UNKNOWN - please submit regex item to check")
                    exit(RCunknown)
                else:
                    for counter in range(len(argvalues)):
                        if len(re.findall(item, argvalues[counter]["text"], re.IGNORECASE)) > 0:
                            errentries=errentries+"Time: %s, State %s, Text: %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(argvalues[counter]["time"])), argvalues[counter]["state"], argvalues[counter]["text"])

                    if errentries == "" :
                        print("OK - pattern %s does not match syslog entries" % (item))
                        RC=RCok
                    else:
                        print("CRITICAL - pattern %s matches syslog entries: %s ......." % (item, errentries[:4000].replace('\n',';')))
                        print(errentries)
                        RC=RCcritical

            else:
                for counter in range(len(argvalues)):
                    if argvalues[counter]["state"] == "SAPControl-YELLOW":
                        errentrieswarn=errentrieswarn+"Time: %s, State %s, Text: %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(argvalues[counter]["time"])), argvalues[counter]["state"], argvalues[counter]["text"])
                    if argvalues[counter]["state"] == "SAPControl-RED":
                        errentriescrit=errentriescrit+"Time: %s, State %s, Text: %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(argvalues[counter]["time"])), argvalues[counter]["state"], argvalues[counter]["text"])

                errentries=errentriescrit+errentrieswarn

                if errentries == "":
                    print("OK - no warnings or errors found")
                    RC=RCok
                elif errentriescrit != "":
                    print("CRITICAL - critical errors found: %s ......." % (errentries[:4000].replace('\n',';')))
                    print(errentries)
                    RC=RCcritical
                else:
                    print("WARNING - warning errors found: %s ......." % (errentries[:4000].replace('\n',';')))
                    print(errentries)
                    RC=RCwarning


        exit(RC)





#Execute selected Mode
if mode == "gettree":
        GetAlertTree()
elif mode == "checkitem":
        if item is None:
                print("UNKNOWN - please submit item to check")
                exit(RCunknown)
        else:
                NagiosOut(GetValue(item))
elif mode == "workprocess":
    NagiosOut(ABAPGetWPTable(item))
elif mode == "getsyslog":
    ABAPReadSyslog()
elif mode == "checksyslog":
    NagiosOut(GetSyslogValue())
else:
        print("UNKNOWN - unknown mode %s" % (mode))
        exit(RCunknown)
