#!/bin/bash

oldpwd=$(pwd)

echo "This script builds and installs the sap plugin.
It requires the sap nw rfc sdk installed at: /usr/sap/.

# run as root
mkdir -p /usr/sap/
tar xf sap-nwrfcsdk.tar.gz
cp -r nwrfcsdk /usr/sap/
"

if [ ! -d /usr/sap/nwrfcsdk ]; then
	echo "Could not find nwrfcsdk. Exit."
	exit 1
fi

tmp_dir=$(mktemp -d -t sapnwbuild-XXXXXXXXXX)

trap "rm -rf $tmp_dir" EXIT INT KILL

set -xe

cp -rpv /opt/openitc/src/openitcockpit-sap/sapnwrfc $tmp_dir/sapnwrfc
cp -rpv /opt/openitc/src/openitcockpit-sap/check_sap_health $tmp_dir/check_sap_health

export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/usr/sap/nwrfcsdk/lib

cd $tmp_dir/sapnwrfc
perl Makefile.PL --source /usr/sap/nwrfcsdk
make
make install

cd $tmp_dir/check_sap_health
./configure && make all
cp plugins-scripts/check_sap_health /opt/openitc/nagios/libexec

cd $oldpwd

echo "Installation finished"

